@@ -27,6 +27,7 @@
 #include "eventsinks.h"
 #include "attachment.h"
 #include "mapihelp.h"
+#include "mimemaker.h"
 #include "message.h"
 #include "revert.h"
 #include "gpgoladdin.h"
@@ -258,10 +259,13 @@ Mail::pre_process_message ()
      to the same value again that it already has) causes
      Outlook to reconsider what it "knows" about a message
      and reread data from the underlying base message. */
-  mapi_change_message_class (message, 1);
-  /* TODO: Unify this so mapi_change_message_class returns
-     a useful value already. */
-  m_type = mapi_get_message_type (message);
+  mapi_change_message_class (message, 1, &m_type);
+
+  if (m_type == MSGTYPE_UNKNOWN)
+    {
+      gpgol_release (message);
+      return 0;
+    }
 
   /* Create moss attachments here so that they are properly
      hidden when the item is read into the model. */
@@ -291,6 +295,13 @@ get_attachment (LPDISPATCH mailitem, int pos)
 
   std::string item_str;
   int count = get_oom_int (attachments, "Count");
+  if (count < 1)
+    {
+      log_debug ("%s:%s: Invalid attachment count: %i.",
+                 SRCNAME, __func__, count);
+      gpgol_release (attachments);
+      return NULL;
+    }
   if (pos > 0)
     {
       item_str = std::string("Item(") + std::to_string(pos) + ")";
@@ -299,13 +310,6 @@ get_attachment (LPDISPATCH mailitem, int pos)
     {
       item_str = std::string("Item(") + std::to_string(count) + ")";
     }
-  if (count < 1)
-    {
-      log_debug ("%s:%s: Invalid attachment count: %i.",
-                 SRCNAME, __func__, count);
-      gpgol_release (attachments);
-      return NULL;
-    }
   attachment = get_oom_object (attachments, item_str.c_str());
   gpgol_release (attachments);
 
@@ -406,9 +410,36 @@ get_attachment_stream (LPDISPATCH mailitem, int pos)
       return NULL;
     }
   LPDISPATCH attachment = get_attachment (mailitem, pos);
-  LPATTACH mapi_attachment = NULL;
   LPSTREAM stream = NULL;
 
+  if (!attachment)
+    {
+      // For opened messages that have ms-tnef type we
+      // create the moss attachment but don't find it
+      // in the OOM. Try to find it through MAPI.
+      HRESULT hr;
+      log_debug ("%s:%s: Failed to find MOSS Attachment. "
+                 "Fallback to MAPI.", SRCNAME, __func__);
+      LPMESSAGE message = get_oom_message (mailitem);
+      if (!message)
+        {
+          log_debug ("%s:%s: Failed to get MAPI Interface.",
+                     SRCNAME, __func__);
+          return NULL;
+        }
+      hr = message->OpenProperty (PR_BODY_A, &IID_IStream, 0, 0,
+                                  (LPUNKNOWN*)&stream);
+      if (hr)
+        {
+          log_debug ("%s:%s: OpenProperty failed: hr=%#lx",
+                     SRCNAME, __func__, hr);
+          return NULL;
+        }
+      return stream;
+    }
+
+  LPATTACH mapi_attachment = NULL;
+
   mapi_attachment = (LPATTACH) get_oom_iunknown (attachment,
                                                  "MapiObject");
   gpgol_release (attachment);
@@ -559,7 +590,7 @@ copy_attachment_to_file (std::shared_ptr<Attachment> att, HANDLE hFile)
   return 0;
 }
 
-/** Sets some meta data on the last attachment atted. The meta
+/** Sets some meta data on the last attachment added. The meta
   data is taken from the attachment object. */
 static int
 fixup_last_attachment (LPDISPATCH mail, std::shared_ptr<Attachment> attachment)
@@ -662,6 +693,14 @@ do_parsing (LPVOID arg)
   log_debug ("%s:%s: preparing the parser for: %p",
              SRCNAME, __func__, arg);
 
+  if (!Mail::is_valid_ptr (mail))
+    {
+      log_debug ("%s:%s: cancel for: %p already deleted",
+                 SRCNAME, __func__, arg);
+      gpgrt_lock_unlock (&parser_lock);
+      return 0;
+    }
+
   if (!parser)
     {
       log_error ("%s:%s: no parser found for mail: %p",
@@ -717,6 +756,7 @@ Mail::decrypt_verify()
 
   if (opt.prefer_html)
     {
+      m_orig_body = get_oom_string (m_mailitem, "HTMLBody");
       if (put_oom_string (m_mailitem, "HTMLBody", placeholder_buf))
         {
           log_error ("%s:%s: Failed to modify html body of item.",
@@ -725,6 +765,7 @@ Mail::decrypt_verify()
     }
   else
     {
+      m_orig_body = get_oom_string (m_mailitem, "Body");
       if (put_oom_string (m_mailitem, "Body", placeholder_buf))
         {
           log_error ("%s:%s: Failed to modify body of item.",
@@ -798,6 +839,30 @@ Mail::update_body()
         }
       return;
     }
+  if (m_verify_result.error())
+    {
+      log_error ("%s:%s: Verification failed. Restoring Body.",
+                 SRCNAME, __func__);
+      if (opt.prefer_html)
+        {
+          if (put_oom_string (m_mailitem, "HTMLBody", m_orig_body.c_str ()))
+            {
+              log_error ("%s:%s: Failed to modify html body of item.",
+                         SRCNAME, __func__);
+            }
+        }
+      else
+        {
+          if (put_oom_string (m_mailitem, "Body", m_orig_body.c_str ()))
+            {
+              log_error ("%s:%s: Failed to modify html body of item.",
+                         SRCNAME, __func__);
+            }
+        }
+      return;
+    }
+  // No need to carry body anymore
+  m_orig_body = std::string();
   auto html = m_parser->get_html_body ();
   /** Outlook does not show newlines if \r\r\n is a newline. We replace
     these as apparently some other buggy MUA sends this. */
@@ -922,6 +987,56 @@ Mail::encrypt_sign ()
 
   const auto window = get_active_hwnd ();
 
+  if (m_is_gsuite)
+    {
+      auto att_table = mapi_create_attach_table (message, 0);
+      int n_att_usable = count_usable_attachments (att_table);
+      mapi_release_attach_table (att_table);
+      /* Check for attachments if we have some abort. */
+      if (n_att_usable)
+        {
+          MessageBox (window,
+                      _("G Suite Sync breaks outgoing crypto mails "
+                        "with attachments.\nUsing crypto and attachments "
+                        "with G Suite Sync is not supported.\n\n"
+                        "See: https://dev.gnupg.org/T3545 for details."),
+                      _("GpgOL: Oops, G Suite Sync account detected"),
+                      MB_ICONINFORMATION|MB_OK);
+          return -1;
+        }
+      if (flags == 2)
+        {
+          MessageBox (window,
+                      _("G Suite Sync breaks outgoing signed mails.\n"
+                        "Ensuring mail integrity (signing) with G Suite Sync "
+                        "is not supported.\n\n"
+                        "See: https://dev.gnupg.org/T3545 for details."),
+                      _("GpgOL: Oops, G Suite Sync account detected"),
+                      MB_ICONINFORMATION|MB_OK);
+          return -1;
+        }
+      if (flags == 3)
+        {
+          if(MessageBox (window,
+                        _("G Suite Sync breaks outgoing signed mails.\n"
+                          "Ensuring mail integrity (signing) with G Suite Sync "
+                          "is not supported.\n\n"
+                          "See: https://dev.gnupg.org/T3545 for details.\n\n"
+                          "Do you want to only encrypt the message?"),
+                        _("GpgOL: Oops, G Suite Sync account detected"),
+                        MB_ICONINFORMATION|MB_YESNO) != IDYES)
+            {
+              return -1;
+            }
+          else
+            {
+              flags = 1;
+            }
+        }
+    }
+
+  m_do_inline = m_is_gsuite ? true : opt.inline_pgp;
+
   EnableWindow (window, FALSE);
   if (flags == 3)
     {
@@ -1005,8 +1120,15 @@ Mail::update_oom_data ()
 
   /* Store the body. It was not obvious for me (aheinecke) how
      to access this through MAPI. */
-  m_html_body = get_oom_string (m_mailitem, "HTMLBody");
-
+  if (m_is_html_alternative)
+    {
+      log_debug ("%s:%s: Is html alternative mail.", SRCNAME, __func__);
+      const char * html_body = get_oom_string (m_mailitem, "HTMLBody");
+      if (html_body)
+        {
+          m_html_body = html_body;
+        }
+    }
   /* For some reason outlook may store the recipient address
      in the send using account field. If we have SMTP we prefer
      the SenderEmailAddress string. */
@@ -1024,6 +1146,7 @@ Mail::update_oom_data ()
           char *senderMail = get_oom_string (m_mailitem, "SenderEmailAddress");
           if (senderMail)
             {
+              log_debug ("%s:%s Sender found", SRCNAME, __func__);
               m_sender = senderMail;
               xfree (senderMail);
               xfree (type);
@@ -1036,20 +1159,38 @@ Mail::update_oom_data ()
   if (sender)
     {
       char *buf = get_oom_string (sender, "SmtpAddress");
-      m_sender = buf;
-      xfree (buf);
+      char *dispName = get_oom_string (sender, "DisplayName");
       gpgol_release (sender);
-      return 0;
+
+      /* Check for G Suite account */
+      if (dispName && !strcmp ("G Suite", dispName))
+        {
+          m_is_gsuite = true;
+        }
+      xfree (dispName);
+      if (buf && strlen (buf))
+        {
+          log_debug ("%s:%s Sender fallback 1", SRCNAME, __func__);
+          m_sender = buf;
+          xfree (buf);
+          return 0;
+        }
+      xfree (buf);
     }
   /* Fallback to Sender object */
   sender = get_oom_object (m_mailitem, "Sender");
   if (sender)
     {
       char *buf = get_pa_string (sender, PR_SMTP_ADDRESS_DASL);
-      m_sender = buf;
-      xfree (buf);
       gpgol_release (sender);
-      return 0;
+      if (buf && strlen (buf))
+        {
+          log_debug ("%s:%s Sender fallback 2", SRCNAME, __func__);
+          m_sender = buf;
+          xfree (buf);
+          return 0;
+        }
+      xfree (buf);
     }
   /* We don't have s sender object or SendUsingAccount,
      well, in that case fall back to the current user. */
@@ -1057,10 +1198,15 @@ Mail::update_oom_data ()
   if (sender)
     {
       char *buf = get_pa_string (sender, PR_SMTP_ADDRESS_DASL);
-      m_sender = buf;
-      xfree (buf);
       gpgol_release (sender);
-      return 0;
+      if (buf && strlen (buf))
+        {
+          log_debug ("%s:%s Sender fallback 3", SRCNAME, __func__);
+          m_sender = buf;
+          xfree (buf);
+          return 0;
+        }
+      xfree (buf);
     }
 
   log_debug ("%s:%s: All fallbacks failed.",
@@ -1085,6 +1231,16 @@ Mail::close_all_mails ()
   std::map<LPDISPATCH, Mail *> mail_map_copy = g_mail_map;
   for (it = mail_map_copy.begin(); it != mail_map_copy.end(); ++it)
     {
+      /* XXX For non racy code the is_valid_ptr check should not
+         be necessary but we crashed sometimes closing a destroyed
+         mail. */
+      if (!is_valid_ptr (it->second))
+        {
+          log_debug ("%s:%s: Already deleted mail for %p",
+                   SRCNAME, __func__, it->first);
+          continue;
+        }
+
       if (!it->second->is_crypto_mail())
         {
           continue;
@@ -1311,7 +1467,8 @@ Mail::close (Mail *mail)
   int rc = invoke_oom_method_with_parms (mail->item(), "Close",
                                        NULL, &dispparams);
 
-  log_debug ("returned from invoke");
+  log_oom_extra ("%s:%s: Reurned from close",
+                 SRCNAME, __func__);
   return rc;
 }
 
@@ -1768,13 +1925,26 @@ std::string
 Mail::get_crypto_details()
 {
   std::string message;
-  /* Handle encrypt only */
+
+  /* No signature with keys but error */
+  if (!is_encrypted() && !is_signed () && m_verify_result.error())
+    {
+      message = _("You cannot be sure who sent, "
+                  "modified and read the message in transit.");
+      message += "\n\n";
+      message += _("The message was signed but the verification failed with:");
+      message += "\n";
+      message += m_verify_result.error().asString();
+      return message;
+    }
+  /* No crypo, what are we doing here? */
   if (!is_encrypted () && !is_signed ())
     {
       return _("You cannot be sure who sent, "
                "modified and read the message in transit.");
     }
-  else if (is_encrypted() && !is_signed ())
+  /* Handle encrypt only */
+  if (is_encrypted() && !is_signed ())
     {
       if (in_de_vs_mode ())
        {
